(* Title:      Prim's Minimum Spanning Tree Algorithm
   Author:     Walter Guttmann
   Maintainer: Walter Guttmann <walter.guttmann at canterbury.ac.nz>
*)

section {* Prim's Minimum Spanning Tree Algorithm *}

theory Prim_Minimum_Spanning_Tree

imports "~~/src/HOL/Hoare/Hoare_Logic" Aggregation_Algebras

begin

no_notation
  trancl ("(_\<^sup>+)" [1000] 999)

context m_kleene_algebra
begin

abbreviation "component g r \<equiv> r\<^sup>T * (--g)\<^sup>\<star>"
definition "spanning_tree t g r \<equiv> forest t \<and> t \<le> (component g r)\<^sup>T * (component g r) \<sqinter> --g \<and> component g r \<le> r\<^sup>T * t\<^sup>\<star> \<and> regular t"
definition "minimum_spanning_tree t g r \<equiv> spanning_tree t g r \<and> (\<forall>u . spanning_tree u g r \<longrightarrow> sum (t \<sqinter> g) \<le> sum (u \<sqinter> g))"
definition "precondition g r \<equiv> g = g\<^sup>T \<and> injective r \<and> vector r \<and> regular r"
definition "invariant t v g r \<equiv> precondition g r \<and> v\<^sup>T = r\<^sup>T * t\<^sup>\<star> \<and> spanning_tree t (v * v\<^sup>T \<sqinter> g) r \<and> (\<exists>w . minimum_spanning_tree w g r \<and> t \<le> w)"

lemma span_tree_split:
  assumes "vector r"
    shows "t \<le> (component g r)\<^sup>T * (component g r) \<sqinter> --g \<longleftrightarrow> (t \<le> (component g r)\<^sup>T \<and> t \<le> component g r \<and> t \<le> --g)"
proof -
  have "(component g r)\<^sup>T * (component g r) = (component g r)\<^sup>T \<sqinter> component g r"
    by (metis assms conv_involutive covector_mult_closed vector_conv_covector vector_covector)
  thus ?thesis
    by auto
qed

lemma span_tree_component:
  assumes "spanning_tree t g r"
    shows "component g r = component t r"
  using assms by (simp add: antisym mult_right_isotone star_isotone spanning_tree_def)

text {*
Prim's minimum spanning tree algorithm is correct.
*}

theorem correctness:
  "VARS g r t v e
  { precondition g r \<and> (\<exists>w . minimum_spanning_tree w g r) }
  t := bot;
  v := r;
  WHILE v * -v\<^sup>T \<sqinter> g \<noteq> bot
    INV { invariant t v g r }
     DO e := minatom (v * -v\<^sup>T \<sqinter> g);
        t := t \<squnion> e;
        v := v \<squnion> e\<^sup>T * top
     OD
  { minimum_spanning_tree t g r }"
proof (vcg_simp)
  fix g r
  let ?ss = "r * r\<^sup>T \<sqinter> g"
  assume 1: "precondition g r \<and> (\<exists>w . minimum_spanning_tree w g r)"
  show "invariant bot r g r"
  proof (unfold invariant_def, intro conjI)
    show "precondition g r"
      using 1 by simp
  next
    show "r\<^sup>T = r\<^sup>T * bot\<^sup>\<star>"
      by (simp add: star_absorb)
  next
    show "spanning_tree bot ?ss r"
    proof (unfold spanning_tree_def, intro conjI)
      show "injective bot"
        by simp
    next
      show "acyclic bot"
        by simp
    next
      show "bot \<le> (component ?ss r)\<^sup>T * (component ?ss r) \<sqinter> --?ss"
        by simp
    next
      show "component ?ss r \<le> r\<^sup>T * bot\<^sup>\<star>"
      proof -
        have "component ?ss r \<le> component (r * r\<^sup>T) r"
          by (simp add: mult_right_isotone star_isotone)
        also have "... \<le> r\<^sup>T * 1\<^sup>\<star>"
          using 1 by (metis inf.eq_iff p_antitone regular_one_closed star_sub_one precondition_def)
        also have "... = r\<^sup>T * bot\<^sup>\<star>"
          by (simp add: star.circ_zero star_one)
        finally show ?thesis
          .
      qed
    next
      show "regular bot"
        by simp
    qed
  next
    show "(\<exists>w . minimum_spanning_tree w g r \<and> bot \<le> w)"
      using 1 by simp
  qed
next
  fix g r t v
  let ?vcv = "v * -v\<^sup>T \<sqinter> g"
  let ?vv = "v * v\<^sup>T \<sqinter> g"
  let ?e = "minatom ?vcv"
  let ?c = "component g r"
  let ?g = "--g"
  assume 2: "invariant t v g r \<and> ?vcv \<noteq> bot"
  have 3: "regular v \<and> regular (v * v\<^sup>T) \<and> regular ((r\<^sup>T * ?g\<^sup>\<star>)\<^sup>T * (r\<^sup>T * ?g\<^sup>\<star>)) \<and> regular ((v \<squnion> ?e\<^sup>T * top) * (v \<squnion> ?e\<^sup>T * top)\<^sup>T)"
    using 2 by (metis invariant_def spanning_tree_def precondition_def regular_conv_closed regular_closed_star regular_mult_closed conv_involutive regular_closed_pp regular_closed_top regular_closed_sup minatom_regular)
  have 4: "t \<le> v * v\<^sup>T \<sqinter> ?g"
    using 2 3 by (metis invariant_def spanning_tree_def inf_pp_commute inf.boundedE)
  hence 5: "t \<le> v * v\<^sup>T"
    by simp
  have 6: "t \<le> ?g"
    using 4 by simp
  have 7: "?e \<le> v * -v\<^sup>T \<sqinter> ?g"
    using 3 by (metis minatom_below pp_dist_inf regular_mult_closed regular_closed_p)
  hence 8: "?e \<le> v * -v\<^sup>T"
    by simp
  have 9: "vector v"
    using 2 invariant_def precondition_def by (simp add: covector_mult_closed vector_conv_covector)
  have 10: "?e * t = bot"
    using 5 8 9 et(1) by blast
  have 11: "?e * t\<^sup>T = bot"
    using 5 8 9 et(2) by simp
  have 12: "atom ?e"
    using 2 minatom_atom by simp
  have "r\<^sup>T \<le> r\<^sup>T * t\<^sup>\<star>"
    by (metis mult_right_isotone order_refl semiring.mult_not_zero star.circ_separate_mult_1 star_absorb)
  hence 13: "r\<^sup>T \<le> v\<^sup>T"
    using 2 by (simp add: invariant_def)
  from 2 obtain w where 14: "minimum_spanning_tree w g r \<and> t \<le> w"
    by (metis invariant_def)
  have 15: "vector r \<and> injective r \<and> v\<^sup>T = r\<^sup>T * t\<^sup>\<star> \<and> forest w \<and> t \<le> w \<and> w \<le> (r\<^sup>T * ?g\<^sup>\<star>)\<^sup>T * (r\<^sup>T * ?g\<^sup>\<star>) \<sqinter> ?g \<and> r\<^sup>T * ((r\<^sup>T * ?g\<^sup>\<star>)\<^sup>T * (r\<^sup>T * ?g\<^sup>\<star>) \<sqinter> ?g)\<^sup>\<star> \<le> r\<^sup>T * w\<^sup>\<star>"
    using 2 3 14 invariant_def precondition_def minimum_spanning_tree_def spanning_tree_def reachable_restrict by auto
  hence 16: "w * v \<le> v"
    using predecessors_reachable reachable_restrict by fastforce
  have 17: "g = g\<^sup>T"
    using 2 invariant_def precondition_def by simp
  hence 18: "?g\<^sup>T = ?g"
    using conv_complement by simp
  show "invariant (t \<squnion> ?e) (v \<squnion> ?e\<^sup>T * top) g r"
  proof (unfold invariant_def, intro conjI)
    show "precondition g r"
      using 2 invariant_def by simp
  next
    show "(v \<squnion> ?e\<^sup>T * top)\<^sup>T = r\<^sup>T * (t \<squnion> ?e)\<^sup>\<star>"
      using 2 8 9 10 by (simp add: reachable_inv invariant_def precondition_def spanning_tree_def)
  next
    let ?v = "v \<squnion> ?e\<^sup>T * top"
    let ?G = "?v * ?v\<^sup>T \<sqinter> g"
    show "spanning_tree (t \<squnion> ?e) ?G r"
    proof (unfold spanning_tree_def, intro conjI)
      show "injective (t \<squnion> ?e)"
        using 2 11 12 by (simp add: injective_inv invariant_def spanning_tree_def)
    next
      show "acyclic (t \<squnion> ?e)"
        using 2 5 8 9 acyclic_inv invariant_def spanning_tree_def by simp
    next
      show "t \<squnion> ?e \<le> (component ?G r)\<^sup>T * (component ?G r) \<sqinter> --?G"
        using 3 4 7 9 15 subgraph_inv inf_pp_commute mst_subgraph_inv_2 by auto
    next
      show "component ((v \<squnion> ?e\<^sup>T * top) * (v \<squnion> ?e\<^sup>T * top)\<^sup>T \<sqinter> g) r \<le> r\<^sup>T * (t \<squnion> ?e)\<^sup>\<star>"
      proof -
        have 19: "r\<^sup>T * ((v * v\<^sup>T) \<sqinter> ?g)\<^sup>\<star> \<le> r\<^sup>T * t\<^sup>\<star>"
          using 2 3 by (metis invariant_def spanning_tree_def inf_pp_commute)
        have "component ((v \<squnion> ?e\<^sup>T * top) * (v \<squnion> ?e\<^sup>T * top)\<^sup>T \<sqinter> g) r = r\<^sup>T * ((v \<squnion> ?e\<^sup>T * top) * (v \<squnion> ?e\<^sup>T * top)\<^sup>T \<sqinter> ?g)\<^sup>\<star>"
          using 3 by simp
        also have "... \<le> r\<^sup>T * (t \<squnion> ?e)\<^sup>\<star>"
          using 4 8 9 12 13 15 18 19 by (metis span_inv)
        finally show ?thesis
          .
      qed
    next
      show "regular (t \<squnion> ?e)"
        using 2 by (metis invariant_def spanning_tree_def regular_closed_sup minatom_regular)
    qed
  next
    show "\<exists>w . minimum_spanning_tree w g r \<and> t \<squnion> ?e \<le> w"
    proof
      let ?f = "w \<sqinter> v * -v\<^sup>T \<sqinter> top * ?e * w\<^sup>T\<^sup>\<star>"
      let ?p = "w \<sqinter> -v * -v\<^sup>T \<sqinter> top * ?e * w\<^sup>T\<^sup>\<star>"
      let ?fp = "w \<sqinter> -v\<^sup>T \<sqinter> top * ?e * w\<^sup>T\<^sup>\<star>"
      let ?w = "(w \<sqinter> -?fp) \<squnion> ?p\<^sup>T \<squnion> ?e"
      have 20: "regular ?f \<and> regular ?fp \<and> regular ?w"
        using 3 14 by (metis regular_conv_closed regular_closed_star regular_mult_closed regular_closed_top regular_closed_inf regular_closed_sup minatom_regular minimum_spanning_tree_def spanning_tree_def)
      show "minimum_spanning_tree ?w g r \<and> t \<squnion> ?e \<le> ?w"
      proof (intro conjI)
        show "minimum_spanning_tree ?w g r"
        proof (unfold minimum_spanning_tree_def, intro conjI)
          show "spanning_tree ?w g r"
          proof (unfold spanning_tree_def, intro conjI)
            show "injective ?w"
              using 8 9 12 15 exchange_injective by blast
          next
            show "acyclic ?w"
              using 8 9 15 16 exchange_acyclic by blast
          next
            show "?w \<le> ?c\<^sup>T * ?c \<sqinter> --g"
            proof -
              have 21: "w \<sqinter> -?fp \<le> ?c\<^sup>T * ?c \<sqinter> --g"
                using 14 by (simp add: le_infI1 minimum_spanning_tree_def spanning_tree_def)
              have "?p\<^sup>T \<le> w\<^sup>T"
                by (simp add: conv_isotone inf.sup_monoid.add_assoc)
              also have "... \<le> (?c\<^sup>T * ?c \<sqinter> --g)\<^sup>T"
                using 15 conv_order by simp
              also have "... = ?c\<^sup>T * ?c \<sqinter> --g"
                using 3 18 conv_dist_comp conv_dist_inf by simp
              finally have 22: "?p\<^sup>T \<le> ?c\<^sup>T * ?c \<sqinter> --g"
                .
              have "?e \<le> (r\<^sup>T * ?g\<^sup>\<star>)\<^sup>T * (r\<^sup>T * ?g\<^sup>\<star>) \<sqinter> ?g"
                using 6 7 15 mst_subgraph_inv by auto
              thus ?thesis
                using 21 22 by simp
            qed
          next
            show "?c \<le> r\<^sup>T * ?w\<^sup>\<star>"
            proof -
              have "?c \<le> r\<^sup>T * w\<^sup>\<star>"
                using 14 minimum_spanning_tree_def spanning_tree_def by simp
              also have "... \<le> r\<^sup>T * ?w\<^sup>\<star>"
                using 5 8 9 14 15 16 20 by (metis mst_reachable_inv)
              finally show ?thesis
                .
            qed
          next
            show "regular ?w"
              using 20 by simp
          qed
        next
          have 23: "?f \<squnion> ?p = ?fp"
            using 3 9 epm_1 by fastforce
          have "atom (w \<sqinter> --v * -v\<^sup>T \<sqinter> top * ?e * w\<^sup>T\<^sup>\<star>)"
            using 6 7 9 12 15 16 reachable_restrict atom_edge by auto
          hence 24: "atom ?f"
            using 3 by simp
          hence "?f = bot \<longrightarrow> top = bot"
            by (metis mult_left_zero mult_right_zero)
          hence "?f \<noteq> bot"
            using 2 le_bot by auto
          hence "?f \<sqinter> v * -v\<^sup>T \<sqinter> ?g \<noteq> bot"
            using 3 15 by (simp add: inf.absorb1 le_infI1)
          hence "g \<sqinter> (?f \<sqinter> v * -v\<^sup>T) \<noteq> bot"
            using inf_commute pp_inf_bot_iff by auto
          hence 25: "?f \<sqinter> ?vcv \<noteq> bot"
            by (simp add: inf_assoc inf_commute)
          hence 26: "?f \<sqinter> g = ?f \<sqinter> ?vcv"
            using 3 by (simp add: inf_assoc inf_commute inf_left_commute)
          have 27: "?e \<sqinter> g = minatom ?vcv \<sqinter> ?vcv"
            using 8 by (simp add: inf.absorb2 inf.assoc inf.commute)
          hence 28: "sum (?e \<sqinter> g) \<le> sum (?f \<sqinter> g)"
            using 20 24 25 26 by (simp add: minatom_min)
          have "?e \<noteq> bot"
            using 25 comp_inf.semiring.mult_not_zero semiring.mult_not_zero by blast
          hence 29: "?e \<sqinter> g \<noteq> bot"
            using 27 minatom_meet_bot by auto
          have "sum (?w \<sqinter> g) = sum (w \<sqinter> -?fp \<sqinter> g) + sum (?p\<^sup>T \<sqinter> g) + sum (?e \<sqinter> g)"
            using 8 9 14 by (metis sum_disjoint_3 epm_8 epm_9 epm_10 minimum_spanning_tree_def spanning_tree_def)
          also have "... \<le> sum (w \<sqinter> -?fp \<sqinter> g) + sum (?p \<sqinter> g) + sum (?f \<sqinter> g)"
            using 17 28 29 by (metis sum_plus_right_isotone sum_symmetric)
          also have "... = sum (((w \<sqinter> -?fp) \<squnion> ?p \<squnion> ?f) \<sqinter> g)"
            using 3 9 by (metis sum_disjoint_3 epm_11 epm_12 epm_13)
          also have "... = sum (w \<sqinter> g)"
            using 3 9 20 23 epm_2 by force
          finally have "sum (?w \<sqinter> g) \<le> sum (w \<sqinter> g)"
            .
          thus "\<forall>u . spanning_tree u g r \<longrightarrow> sum (?w \<sqinter> g) \<le> sum (u \<sqinter> g)"
            using 14 order_lesseq_imp minimum_spanning_tree_def by auto
        qed
      next
        show "t \<squnion> ?e \<le> ?w"
          using 5 9 14 mst_extends_new_tree by simp
      qed
    qed
  qed
next
  fix g r t v
  let ?g = "--g"
  assume 30: "invariant t v g r \<and> v * -v\<^sup>T \<sqinter> g = bot"
  have 31: "regular v \<and> regular (v * v\<^sup>T)"
    using 30 by (metis invariant_def spanning_tree_def precondition_def regular_conv_closed regular_closed_star regular_mult_closed conv_involutive)
  have 32: "v * -v\<^sup>T \<sqinter> ?g = bot"
    using 30 pp_inf_bot_iff pp_pp_inf_bot_iff by simp
  have 33: "v\<^sup>T = r\<^sup>T * t\<^sup>\<star> \<and> vector v"
    using 30 by (simp add: covector_mult_closed invariant_def vector_conv_covector precondition_def)
  have 34: "t \<le> v * v\<^sup>T \<sqinter> ?g"
    using 30 31 by (metis invariant_def inf_pp_commute spanning_tree_def inf.boundedE)
  have "r\<^sup>T * (v * v\<^sup>T \<sqinter> ?g)\<^sup>\<star> \<le> r\<^sup>T * t\<^sup>\<star>"
    using 30 31 by (metis invariant_def inf_pp_commute spanning_tree_def)
  hence 35: "component g r = v\<^sup>T"
    using 31 32 33 34 by (metis span_post)
  from 30 obtain w where 36: "minimum_spanning_tree w g r \<and> t \<le> w"
    by (metis invariant_def)
  have 37: "w \<le> v * v\<^sup>T"
    using 31 35 36 by (simp add: minimum_spanning_tree_def spanning_tree_def inf_pp_commute)
  have "vector r \<and> injective r \<and> forest w"
    using 30 36 by (simp add: invariant_def precondition_def minimum_spanning_tree_def spanning_tree_def)
  hence "w = t"
    using 33 36 37 mst_post by blast
  thus "minimum_spanning_tree t g r"
    using 36 by simp
qed

end

end

